@@ -6,11 +6,16 @@ from os import environ
 from pathlib import Path
 from subprocess import DEVNULL, PIPE, CalledProcessError
 from subprocess import run as run
+from textwrap import wrap
 
 import click
+import toml
 from temppathlib import TemporaryDirectory
 from tqdm import tqdm
 
+# TODO
+# * shallow-update
+
 batch = dict(environ)
 batch["VISUAL"] = "/bin/true"
 
@@ -24,9 +29,13 @@ def fork(channel):
     run(["pijul", "fork", channel], check=True)
 
 
+def fork_internal(channel):
+    fork(f"in_{channel}")
+
+
 def git_restore():
     run(["git", "checkout", "--no-overlay", "-q", "."], check=True)
-    run(["git", "clean", "-xdfq", "-e", ".pijul", "-e", ".ignore"], check=True)
+    run(["git", "clean", "-xdfq", "-e", ".pijul"], check=True)
 
 
 def pijul_restore():
@@ -41,29 +50,16 @@ def switch(channel):
     run(["pijul", "channel", "switch", channel], check=True, stdout=DEVNULL)
 
 
+def switch_internal(channel):
+    switch(f"in_{channel}")
+
+
 def delete(channel):
     run(["pijul", "channel", "delete", channel], check=True, stdout=DEVNULL)
 
 
-def ancestry_path(head, base):
-    res = run(
-        ["git", "rev-list", "--ancestry-path", f"{base}..{head}"],
-        check=True,
-        stdout=PIPE,
-    )
-    return res.stdout.splitlines()
-
-
-def get_channels():
-    return run(["pijul", "channel"], check=True, stdout=PIPE).stdout.decode("UTF-8")
-
-
-def get_head():
-    res = run(["git", "rev-parse", "HEAD"], check=True, stdout=PIPE)
-    head = res.stdout.strip().decode("UTF-")
-    res = run(["git", "rev-parse", "--abbrev-ref", "HEAD"], check=True, stdout=PIPE)
-    name = res.stdout.strip().decode("UTF-")
-    return head, name
+def delete_internal(channel):
+    delete(f"in_{channel}")
 
 
 def clone(git_repo):
@@ -75,7 +71,9 @@ def checkout(rev):
 
 
 def add_recursive():
-    run(["pijul", "add", "-r", "."], check=True)
+    for item in Path(".").iterdir():
+        if item.name not in [".pijul", ".git", "."]:
+            run(["pijul", "add", "-r", bytes(item)], check=True)
 
 
 def record(log, author, timestamp):
@@ -116,6 +114,40 @@ def record_simple(log):
     return res.stdout.strip().decode("UTF-8"), res.stderr.strip().decode("UTF-8")
 
 
+def get_changes(channel=None):
+    cmd = ["pijul", "log", "--hash-only"]
+    if channel:
+        cmd += ["--channel", channel]
+    res = run(cmd, check=True, stdout=PIPE)
+    return set(res.stdout.decode("UTF-8").splitlines())
+
+
+def get_change(hash):
+    res = run(["pijul", "change", hash], check=True, stdout=PIPE)
+    return res.stdout.decode("UTF-8")
+
+
+def get_ancestry_path(head, base):
+    res = run(
+        ["git", "rev-list", "--ancestry-path", f"{base}..{head}"],
+        check=True,
+        stdout=PIPE,
+    )
+    return res.stdout.splitlines()
+
+
+def get_channels():
+    return run(["pijul", "channel"], check=True, stdout=PIPE).stdout.decode("UTF-8")
+
+
+def get_head():
+    res = run(["git", "rev-parse", "HEAD"], check=True, stdout=PIPE)
+    head = res.stdout.strip().decode("UTF-")
+    res = run(["git", "rev-parse", "--abbrev-ref", "HEAD"], check=True, stdout=PIPE)
+    name = res.stdout.strip().decode("UTF-")
+    return head, name
+
+
 def get_base(head):
     return (
         run(
@@ -128,7 +160,7 @@ def get_base(head):
     )
 
 
-def show():
+def get_show():
     return run(["git", "show", "-s"], check=True, stdout=PIPE).stdout.decode(
         "UTF-8", errors="ignore"
     )
@@ -142,6 +174,25 @@ def get_tag(res, tag, line):
     return False
 
 
+def get_rev_list(head, base):
+    return (
+        run(
+            [
+                "git",
+                "rev-list",
+                "--topo-order",
+                "--ancestry-path",
+                "--no-merges",
+                f"{base}..{head}",
+            ],
+            check=True,
+            stdout=PIPE,
+        )
+        .stdout.decode("UTF-8")
+        .splitlines()
+    ) + [base]
+
+
 def parse_date(date):
     return datetime.strptime(" ".join(date.split(" ")[:-1]), "%a %b %d %H:%M:%S %Y")
 
@@ -164,25 +215,6 @@ def parse_log(log):
     return "\n".join(out), res["author"], parse_date(res["date"])
 
 
-def rev_list(head, base):
-    return (
-        run(
-            [
-                "git",
-                "rev-list",
-                "--topo-order",
-                "--ancestry-path",
-                "--no-merges",
-                f"{base}..{head}",
-            ],
-            check=True,
-            stdout=PIPE,
-        )
-        .stdout.decode("UTF-8")
-        .splitlines()
-    ) + [base]
-
-
 def rename(a, b):
     a.parent.mkdir(parents=True, exist_ok=True)
     b.parent.mkdir(parents=True, exist_ok=True)
@@ -194,6 +226,161 @@ def rename(a, b):
         return False
 
 
+def find_current_channel():
+    channels = get_channels().splitlines()
+    for channel in channels:
+        if channel.startswith("* "):
+            _, _, result = channel.partition("* ")
+            return result
+    raise click.UsageError("No current pijul channel??")
+
+
+re_dep = re.compile(r"\d\] ([a-zA-Z0-9]{53})\b")
+
+
+def find_dependencies(not_in=None):
+    dep_dict = dict()
+    changes = get_changes()
+    if not_in:
+        changes -= get_changes(not_in)
+    for change_hash in changes:
+        change = get_change(change_hash)
+        in_block = False
+        dependencies = []
+        toml_lines = []
+        for line in change.splitlines():
+            if in_block:
+                if not line.split():
+                    break
+                split = re_dep.findall(line)
+                if split:
+                    dependencies.append(split[0])
+            elif line.startswith("# Dependencies"):
+                in_block = True
+            elif line.startswith("# Hunks"):
+                break
+            else:
+                toml_lines.append(line)
+        toml_code = os.linesep.join(toml_lines)
+        info = toml.loads(toml_code)
+        info["hash"] = change_hash
+        info["dependencies"] = dependencies
+        dep_dict[change_hash] = info
+    return dep_dict
+
+
+re_rev = re.compile(r"\bin_[A-Fa-f0-9]{40}\b", re.MULTILINE)
+
+
+def find_shortest_path(head):
+    res = []
+    channels = get_channels()
+    for base in re_rev.findall(channels):
+        _, _, base = base.partition("_")
+        length = len(get_ancestry_path(head, base))
+        res.append((length, head, base))
+    res = sorted(res, key=lambda x: x[0])
+    return res[0]
+
+
+re_commit = re.compile(r"\bcommit [A-Fa-f0-9]{40}\b", re.MULTILINE)
+
+
+def extract_subject(msg):
+    for line in msg.splitlines():
+        line = line.strip()
+        if not line or re_commit.match(line):
+            continue
+        line = "\\n".join(wrap(line, 40))
+        return line
+    return "no message"
+
+
+def plot_nodes(deps):
+    res = []
+    for item in deps.values():
+        hash = item["hash"]
+        msg = extract_subject(item["message"])
+        res.append(f'"{hash}" [ label = "{msg}" ];')
+    return os.linesep.join(res)
+
+
+def plot_edges(deps):
+    res = []
+    for item in deps.values():
+        for dep in item["dependencies"]:
+            if dep in deps:
+                start = item["hash"]
+                res.append(f'"{start}" -> "{dep}";')
+    return os.linesep.join(res)
+
+
+def plot_digraph(not_in=None, rank_lr=True):
+    deps = find_dependencies(not_in)
+    if rank_lr:
+        rank = "rankdir=LR;"
+    else:
+        rank = ""
+    return """
+digraph {{
+{rank}
+{nodes}
+{edges}
+}}
+""".format(
+        rank=rank, nodes=plot_nodes(deps), edges=plot_edges(deps)
+    )
+
+
+def prepare_workdir(workdir, tmp_dir):
+    os.chdir(bytes(tmp_dir.path))
+    clone(workdir)
+    Path(".pijul").symlink_to(Path(workdir, ".pijul"))
+
+
+def run_it(head, base):
+    revs = get_rev_list(head, base)
+    runner = Runner(revs)
+    try:
+        runner.run()
+    except:  # noqa
+        if runner.error:
+            delete_internal(head)
+        raise
+
+
+def final_fork(head):
+    head = head[:7]
+    head = f"work_{head}"
+    fork(head)
+    switch(head)
+    print("Please do not work in internal in_* channels\n")
+    print("If you like to rename the new work channel call:\n")
+    print(f"pijul channel rename {head} $new_name")
+
+
+def check_git():
+    if not Path(".git").exists():
+        raise click.UsageError("Please, use git-pijul in root of git-repository")
+
+
+def check_init():
+    if not Path(".pijul").exists():
+        init()
+    else:
+        raise click.UsageError("'.pijul' already exists, please use 'update'")
+
+
+def check_head(head):
+    if head is None:
+        head, name = get_head()
+        if not name:
+            print(f"Using head: {head}")
+        else:
+            print(f"Using head: {head} ({name})")
+    return head
+
+
 class Runner:
     def __init__(self, revs):
         self.revs = revs
@@ -218,16 +405,16 @@ class Runner:
                 prev.append(rev)
                 for last in reversed(prev):
                     if last in info:
-                        fork(last)
+                        fork_internal(last)
                         pijul_restore()
-                        switch(last)
+                        switch_internal(last)
                         self.error = True
                         break
                 raise
 
     def prepare(self, prev, rev):
         checkout(rev)
-        log = show()
+        log = get_show()
         log, author, date = parse_log(log)
         timestamp = str(int(date.timestamp()))
         return log, author, timestamp
@@ -240,65 +427,18 @@ class Runner:
         hash_ = hash_.strip()
 
 
-def check_git():
-    if not Path(".git").exists():
-        raise click.UsageError("Please, use git-pijul in root of git-repository")
-
-
-def check_init():
-    if not Path(".pijul").exists():
-        init()
-    else:
-        raise click.UsageError("'.pijul' already exists, please use 'update'")
-
-
-def check_head(head):
-    if head is None:
-        head, name = get_head()
-        if not name:
-            print(f"Using head: {head}")
-        else:
-            print(f"Using head: {head} ({name})")
-    return head
-
-
-re_rev = re.compile(r"\b[A-Fa-f0-9]{40}\b", re.MULTILINE)
-
-
-def find_shortest_path(head):
-    res = []
-    channels = get_channels()
-    for base in re_rev.findall(channels):
-        length = len(ancestry_path(head, base))
-        res.append((length, head, base))
-    res = sorted(res, key=lambda x: x[0])
-    return res[0]
-
-
-def prepare_workdir(workdir, tmp_dir):
-    os.chdir(bytes(tmp_dir.path))
-    clone(workdir)
-    Path(".pijul").symlink_to(Path(workdir, ".pijul"))
-
-
-def add_ignore():
-    ignore = Path(".ignore").absolute()
-    if not ignore.exists():
-        with open(bytes(ignore), "w") as f:
-            f.truncate(0)
-            f.write(".git\n")
-
-
-def run_it(head, base):
-    add_ignore()
-    revs = rev_list(head, base)
-    runner = Runner(revs)
-    try:
-        runner.run()
-    except:  # noqa
-        if runner.error:
-            delete(head)
-        raise
+def fill_channel_sets(left, right):
+    if not left:
+        left = [find_current_channel()]
+    if not right:
+        right = [find_current_channel()]
+    left_set = set()
+    for item in left:
+        left_set.update(get_changes(item))
+    right_set = set()
+    for item in right:
+        right_set.update(get_changes(item))
+    return left_set, right_set
 
 
 @click.group()
@@ -306,27 +446,85 @@ def main():
     pass
 
 
+@main.command()
+@click.option(
+    "--left", "-l", multiple=True, help="Left channel, multiple options allowed"
+)
+@click.option(
+    "--right", "-r", multiple=True, help="Right channel, multiple options allowed"
+)
+def set_diff(left, right):
+    """Difference between two sets changes of channels. union(left*) \\ union(right*)"""
+    left_set, right_set = fill_channel_sets(left, right)
+    result = left_set - right_set
+    for item in result:
+        print(item)
+
+
+@main.command()
+@click.option(
+    "--left", "-l", multiple=True, help="Left channel, multiple options allowed"
+)
+@click.option(
+    "--right", "-r", multiple=True, help="Right channel, multiple options allowed"
+)
+def set_intersection(left, right):
+    """Intersection between two sets changes of channels. union(left*) \\ union(right*)"""
+    left_set, right_set = fill_channel_sets(left, right)
+    result = left_set & right_set
+    for item in result:
+        print(item)
+
+
+@main.command()
+@click.option(
+    "--channel", "-c", multiple=True, help="Channel, multiple options allowed"
+)
+def set_union(channel):
+    """Union changes of channels. union(channel*)"""
+    channel_set = set()
+    for item in channel:
+        channel_set.update(get_changes(item))
+    for item in channel_set:
+        print(item)
+
+
+@main.command()
+@click.option(
+    "--rank-lr/--no-rank-lr",
+    "-h/-v",
+    default=True,
+    help="Remove changes in than channel",
+)
+@click.option("--not-in", "-i", default=None, help="Remove changes in than channel")
+def plot(not_in, rank_lr):
+    """Display current changes as graphviz file (git pijul plot | dot -Txlib)"""
+    print(plot_digraph(not_in, rank_lr))
+
+
 @main.command()
 def shallow():
-    """create a new pijul repository from current revision without history"""
+    """create a new pijul repository from the current revision without history"""
     workdir = Path(".").absolute()
     check_git()
     check_init()
     with TemporaryDirectory() as tmp_dir:
         prepare_workdir(workdir, tmp_dir)
         head, _ = get_head()
-        fork(head)
-        switch(head)
-        add_ignore()
+        fork_internal(head)
+        switch_internal(head)
         add_recursive()
         record_simple(f"commit {head}")
+        final_fork(head)
 
 
 @main.command()
-@click.option("--base", default=None, help="Update from (commit-ish, default '--root')")
-@click.option("--head", default=None, help="Update to (commit-ish, default HEAD)")
+@click.option(
+    "--base", "-b", default=None, help="Update from (commit-ish, default '--root')"
+)
+@click.option("--head", "-h", default=None, help="Update to (commit-ish, default HEAD)")
 def update(base, head):
-    """update a repository created with git-pijul"""
+    """Update a repository created with git-pijul"""
     workdir = Path(".").absolute()
     check_git()
     if not Path(".pijul").exists():
@@ -345,18 +543,21 @@ def update(base, head):
     with TemporaryDirectory() as tmp_dir:
         prepare_workdir(workdir, tmp_dir)
         pijul_restore()
-        switch(base)
-        fork(head)
-        switch(head)
+        switch_internal(base)
+        fork_internal(head)
+        switch_internal(head)
         git_restore()
         run_it(head, base)
+        final_fork(head)
 
 
 @main.command()
-@click.option("--base", default=None, help="Import from (commit-ish, default '--root')")
-@click.option("--head", default=None, help="Import to (commit-ish, default HEAD)")
+@click.option(
+    "--base", "-b", default=None, help="Import from (commit-ish, default '--root')"
+)
+@click.option("--head", "-h", default=None, help="Import to (commit-ish, default HEAD)")
 def create(base, head):
-    """create a new pijul repository"""
+    """Create a new pijul repository and import a linear history"""
     workdir = Path(".").absolute()
     check_git()
     check_init()
@@ -366,9 +567,10 @@ def create(base, head):
         print(f"Using base: {base} ('--root')")
     with TemporaryDirectory() as tmp_dir:
         prepare_workdir(workdir, tmp_dir)
-        fork(head)
-        switch(head)
+        fork_internal(head)
+        switch_internal(head)
         run_it(head, base)
+        final_fork(head)
 
 
 if __name__ == "__main__":
